# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import math,stats
from score import score

def log(x):
    return math.log(x,2)

class BDE(score):
    def __init__(self,prior=None,*args,**kwds):
        score.__init__(self,*args,**kwds)
        self.prior=prior
        self.alpha = 1.0
        
    def HP(self,v,par):
        if self.prior==None:
            return max(2,v.n_disc)
        else:
            hp=float(self.prior)
            for p in par:
                hp/=max(2,p.n_disc)
            return hp
    
    def H(self,v,par):
        if self.prior==None:
            return 1.0
        else:
            return self.HP(v,par)/max(2,v.n_disc)
    
    def graph_score(self,number_of_parents,gene_vertex,weights_of_parents,number_of_data_points, list_of_parents):
        return sum(map(log,weights_of_parents))*self.alpha*log(number_of_data_points+1)*min(1,gene_vertex.n_disc+0.1)
    
    def lower_bound_for_data_score(self,selected_data_empty):
        if self.prior!=None:
            return 0.0
        HP=self.HP(selected_data_empty.vertex,[])
        H=self.H(selected_data_empty.vertex,[])
        stats_all,stats_par = selected_data_empty.stats()
        if selected_data_empty.vertex.n_disc:
            s = 0
            for a,cv in stats_all.items():
                for i in range(0,cv):
                    s-=log(H+i)
                    s+=log(HP+i)
        else:
            return 0.0
#            gh = stats.gammln(H)
#            s = stats.gammln(HP+stats_par[()])-stats.gammln(HP)
#            for a,cv in stats_all.items():
#                s+=gh-stats.gammln(H+cv)
        return s*self.data_factor
        
    def data_score(self,selected_data):
        HP=self.HP(selected_data.vertex,selected_data.parents)
        H=self.H(selected_data.vertex,selected_data.parents)
        stats_all,stats_par = selected_data.stats()
        s = 0
        if 0 not in map(lambda p: p.n_disc,selected_data.parents+[selected_data.vertex]):
            par_counted = set()
            for a,cv in stats_all.items():
                numbers_of_parents = a[:-1]
                cp = stats_par[numbers_of_parents]
                if numbers_of_parents not in par_counted:
                    tmp =0
                    for i in range(0,cp):
                        tmp+=log(HP+i)
                    s+=tmp
                    par_counted.add(numbers_of_parents)
                for i in range(0,cv):
                    s-=log(H+i)
        else:
            gh = stats.gammln(H)
            ghp = stats.gammln(HP)
            for a,cv in stats_all.items():
                s+=gh-stats.gammln(H+cv)
            for a,cp in stats_par.items():
                s+=stats.gammln(HP+cp)-ghp
        return s*self.data_factor




import random, bisect
from math import exp, log1p, log

def get_weights(dataset, score, **kwargs):
    """Computes weights for all vertices and their potential parents"""
    weights = {}
    for node in dataset.vertices:
        # determine possible parents of the node
        parents = dataset.get_potential_parents(node)
        # compute parents' weights for this node
        pweights = parents_weights(dataset, score, node, parents, **kwargs)
        # add all weights to a dictionary
        for par, w in zip(parents, pweights):
            weights[node, par] = w
    return weights

def parents_weights(dataset, score, node, parents, **kwargs):
    """Weights for given dataset, score function, node and its parents."""
    thres, no_el = thresholds(dataset, score, node, parents, **kwargs)

    weights = [z_k - score.graph_score(len(parents),node,[],no_el,parents) for z_k in thres]
    if any(w < 0.0 for w in weights):
        # print "negative weights !"
        weights = [max(w, 0.0) for w in weights]
    return weights

def thresholds(dataset, score, X, Y, alpha=0.05):
    
    n_parents=len(Y)
    if n_parents < 1:
        return [], 0

    # Determine beta value and threshold values.
    beta = bonferroni_correction(alpha, n_parents)
    # changed from 1e-10 to 1e-5 because now no_tries depends on it
    beta = max(beta, 1e-5)

    no_tries = int(round(max(100, 1.0 / beta)))

    no_el, data = dataset.sel_data(X.index) 
    prs = [shuffled_distribution(data, score, X, Y_k, no_tries) for Y_k in Y]
    models = [cumulative(pr) for pr in prs]

    values = [cm_threshold(beta, model) for model in models]
    return values, no_el

def cm_threshold(beta, cm):
    """
    Return threshold z such that P(Z > z | H0) ~= beta holds.
    """
    xs, ys = zip(*cm)
    # negate values to get proper order
    k = bisect.bisect_left([-y for y in ys], -log(beta))
    if k < len(xs):
        # return (xs[k - 1] + xs[k]) / 2.0
        z = (log(beta) - ys[k - 1]) / (ys[k] - ys[k - 1])
        return xs[k - 1] + (xs[k] - xs[k - 1]) * z
    else:
        return xs[k - 1] + 0.1  # +0.1 is important here !


def bonferroni_correction(alpha, parents_num):
    return alpha / parents_num

# Waszczuk version
def cumulative(points):
    M_INF = float('-inf')

    def log_is_zero(x):
        return x == M_INF

    def log_add(x, y):
        if log_is_zero(x):
            return y
        if x > y:
            return x + log1p(exp(y - x))
        else:
            return y + log1p(exp(x - y))

    def logsum(w):
        """[log(w1), ..., log(wn)] -> log(w1 + ... + wn)"""
        return reduce(log_add, sorted(w), M_INF)

    def partial_sums(l, add, init):
        s = init
        r = []
        for x in l:
            s = add(s, x)
            r.append(s)
        return r

    # flatten
    points.sort()
    old = map(lambda x: (x, 0.0), points)
    points = []
    i = 0
    while i < len(old):
        part = []
        x = old[i][0]
        while i < len(old) and old[i][0] == x:
            part.append(old[i][1])
            i += 1
        points.append((x, logsum(part)))

    # rest
    X = map(lambda x: x[0], points)
    Y = map(lambda x: x[1], points)
    Y = reversed(Y)
    Y = partial_sums(Y, log_add, M_INF)
    maxy = Y[-1]
    Y = [y - maxy for y in Y]
    Y.reverse()
    return zip(X, Y)

def shuffle_data(data, ind, values):
    """Shuffles data from experiments for specified node index (shifted by no of variables)
    parameter data denotes a list of experiments data as in class experiment"""

    random.shuffle(values)

    for i, d in enumerate(data):
        d[ind] = values[i]

def shuffled_distribution(data, score, X, Y, no_tries):
    """Shuffles no_tries times X column and returns z_i values
    we assume, that data is just a copy and can be lost after this operation
    """
    # result distribution as a list of scores
    P = []
    # number of variables / genes
    var_no = len(data[0]) / 2 # due to dynamic/static networks stuff, data in experiments are kept in two copies

    empty_score = score.data_score(X, [], data)
    # compute list of values for variable X
    x_values = [d[X.index + var_no] for d in data]

    # pnum times shuffle X data and add score to the distribution
    for i in range(no_tries):
        shuffle_data(data, X.index + var_no, x_values)
        stat_value = empty_score - score.data_score(X, [Y], data)
        P.append(stat_value)

    return P



# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#




class stack:
    def __init__(self):
        self.stack = []
    def put(self,item):
        self.stack.append(item)
        return self
    def get(self):
        res = self.stack[-1]
        self.stack=self.stack[:-1]
        return res
    def empty(self):
        return len(self.stack)==0

class queue(stack):
    def get(self):
        res = self.stack[0]
        self.stack=self.stack[1:]
        return res


class graph:
    def __init__(self):
        self.vertices = []
        self.edges = {}
        self.edge_labelling = {}
        self.vertice_labelling = {}

    def add_edge(self,st,end,label=" "):
        try:
            self.edges[st].append(end)
        except KeyError:
            if st not in self.vertices:
                raise "Wrong starting vertice"
            self.edges[st] =[end]
            
        self.edge_labelling[st,end]=label
        
    def add_vert(self,vert,label=" "):
        self.vertices.append(vert)
        self.vertice_labelling[vert]=label
        self.edges[vert]=[]
        
    def __repr__(self):
        rep = "Graph: \n"
        for v in self.vertices:
            rep += "\t%s"%v
            try:
                rep += "(%s) => "%str(self.vertice_labelling[v])
            except KeyError:
                rep += " => "
            for v2 in self.edges[v]:
                rep+= "%s"%v2
                try:
                     rep+="(%s), "%str(self.edge_labelling[(v,v2)])
                except KeyError:
                    rep+=", "
            rep +="\n"
        #print self.edge_labelling, self.vertice_labelling
        return rep
            
    def ford_bellman(self,s):
        """
        Bellman - Ford algorithm
        """
        D = {}
        inf = len(self.vertices)
        for x in self.vertices:
            if x in self.edges[s]:
                D[x]=1
            else:
                D[x]=inf
        D[s]=0
        for i in range(len(self.vertices)-2):
            for v in self.vertices:
                if v != s:
                    for u in self.vertices:
                        if v in self.edges[u]:
                            D[v]=min(D[v],D[u]+1)
        return D

    def avg_dist(self):
        s = 0
        for v in self.vertices:
            s+=sum(self.ford_bellman(v).values())
        return s*1.0/(len(self.vertices)**2-len(self.vertices))

    def make_undirected(self):
        for v in self.vertices:
            for w in self.vertices:
                if (w in self.edges[v]) and (not v in self.edges[w]):
                    self.edges[w].append(v)
                    
    def c_v(self,v, method="plain"):
        """clustering coefficient of a vertice"""
        nom = 0
        denom=0
        if method=="plain":
            for u in self.edges[v]: #for all neighbours of v
                for w in self.edges[u]: # for all neighbours of a given neighbour 
                    if w in self.edges[v]: # if it is a neighbour of v
                        nom+=1
            denom=(len(self.edges[v])*(len(self.edges[v])-1)) #divided by possible triangles
        elif method=="FFA": #feed-forward "master"
            for u in self.edges[v]: #for all neighbours of v
                for w in self.edges[u]: # for all neighbours of a given neighbour 
                    if w in self.edges[v]: # if it is a neighbour of v
                        nom+=1
            denom=len(self.edges[v])*(len(self.edges[v])-1) #divided by pairs of outgoing edges
        elif method=="FFB": #feed-forward "helper"
            for u in self.parents(v): #for all parents of v
                for w in self.edges[u]: # for all neighbours of a given parent
                    if w in self.edges[v]: # if it is a neighbour of v
                        nom+=1
            denom=len(self.parents(v))*len(self.edges[v]) #divided  outgoing * incoming edges
        elif method=="FFC": #feed-forward "target"
            for u in self.parents(v): #for all parents of v
                for w in self.edges[u]: # for all neighbours of a given parent
                    if v in self.edges[w]: # if it is a neighbour of v
                        nom+=1
            denom=len(self.parents(v))*(len(self.parents(v))-1) #  incoming edges squared
        elif method=="FB": #feed-back loop
            for u in self.edges[v]: #for all neighbors of v
                for w in self.edges[u]: # for all neighbours of a given neighbor
                    if v in self.edges[w]: # if v is a neighbour of it
                        nom+=1
            denom=len(self.parents(v))*len(self.edges[v]) #divided  outgoing * incoming edges
        
        try:
            return nom*1.0/denom
        except ZeroDivisionError:
            return 0.0
    def clust_sig(self,meths=["FFA","FFB","FFC","FB"]):
        lst=[]
        avg={}
        for m in meths:
            avg[m]=0.0
        for v in self.vertices:
            vec=map(lambda m: self.c_v(v,m),meths)
            s=sum(vec)
            if s==0.0:
                norm_vec=vec
            else:
                norm_vec=map(lambda x: x/s,vec)
            lst.append(norm_vec)
            for m,v in zip(meths,norm_vec):
                avg[m]+=v
        s_all=sum(avg.values())
        for m in meths:
            avg[m]/=s_all
        return lst,avg
    
    def random_regular(self,n,k):
        """
        create a random graph with n vertices, each of degree k
        """
        import random
        self.__init__()
        self.vertices=range(n)
        for v in self.vertices:
            self.edges[v]=random.sample(self.vertices,k)

    def random_of_size(self,n,m,e_labels=["+","-"],v_labels=["AND","OR"]):
        """
        create a random graph with n vertices and m edges
        """
        import random
        self.__init__()
        self.vertices=range(n)
        for v in self.vertices:
            self.vertice_labelling[v]=random.choice(v_labels)
            self.edges[v]=[]
        edge_pool = []
        for u in self.vertices:
            for v in self.vertices:
                edge_pool.append((u,v))

        sample = random.sample(edge_pool,m)

        for u,v in sample:
            self.edges[u].append(v)
            self.edge_labelling[u,v]=random.choice(e_labels)
        
        
        
    def search(self, start, fun, struct=stack,avoid=None,forward=True,backward=False):
        if not type(start)==type([]): # did we get a list of starting points ?
            start = [start]
            
        if avoid==None:
            avoid=[] # We can't set it to [] as default, because static parameter binding sucks !
            
        st = struct()
        for v in start: #put all the starting points on the stack/queue/whatever
            st.put(v)
            avoid.append(v)
            
        while not st.empty():
            vert = st.get()
            fun(vert)
            #print "visiting", vert,self.edges[vert]
            if forward: #do we search along the edge orientation 
                for v in self.edges[vert]:
                    if v not in avoid:
                        st.put(v)
                        avoid.append(v)
            if backward: #do we search "backwards" the edge orientation
                for v in self.parents(vert):
                    if v not in avoid:
                        st.put(v)
                        avoid.append(v)
                    
    def subgraph(self,source,forward=True,backward=False):
        import sys,copy
        res = graph()
        f = lambda v: res.vertices.append(v)#sys.stdout.write("%s\n"%v)
        self.search(source,f,queue,None,forward,backward)
        for v in res.vertices:
            res.vertice_labelling[v]=self.vertice_labelling[v]
            res.edges[v]=copy.copy(self.edges[v])
            for w in res.edges[v]: 
                if not forward: #we need to check if there are no spurious edges
                    if w not in res.vertices:
                        res.edges[v].remove(w)
                        continue
                res.edge_labelling[(v,w)]=self.edge_labelling[(v,w)]
        #pruning
        g=res
        for v in g.vertices:
            for v2 in g.edges[v]:
                if v2 not in g.vertices:
                    g.edges[v].remove(v2)
        return g
        
    def verticeByLabel(self,label):
        for k,v in self.vertice_labelling.items():
            if v==label:
                return k

    def edgeByLabel(self,label):
        for k,v in self.edge_labelling.items():
            if v==label:
                return k

    def toDot(self):
        import pydot
        edges = []
        edg_keys = []
        for g in self.vertices:
            g_lab = str(g) + " - " + (self.vertice_labelling[g])
            for g2 in self.edges[g]:
                g2_lab = str(g2) + " - " + (self.vertice_labelling[g2])
                edges.append((g_lab,g2_lab))
                edg_keys.append((g,g2))  
        #print edges,edg_keys
        g = pydot.graph_from_edges(edges,directed=True)
        for e in g.edge_list:
            i = edges.index((e.get_source(),e.get_destination()))
            if self.edge_labelling[edg_keys[i]]=="+":
                e.set_label("+")
            elif self.edge_labelling[edg_keys[i]]=="-":
                e.set_label("-")
                e.set_arrowhead("tee")
            else: # no info
                e.set_label(self.edge_labelling[edg_keys[i]])
                e.set_arrowhead("none")
        return g

    def show(self):
        import tempfile

        f,fn = tempfile.mkstemp(".jpg")
        #f.close()
        self.toDot().write_jpg(fn)
        import Image
        im = Image.open(fn)
        im.show()
        import os
        os.remove(fn)
    
    def parents(self,vert):
        par =[]
        for v in self.vertices:
            if vert in self.edges[v]:
                par.append(v)
        return par
        
    def fromParents(self,verts,pars):
        self.__init__()
        self.vertices=[]
        for v in verts:
            self.add_vert(v)
            self.edges[v]=[]
            
        for v,par in zip(verts,pars):
            for p in par:
                self.add_edge(p,v)
        return self
    
    def fromOIMfile(self,f):
        ln = f.readline()
        par = []
        while ln and ln.find("-")==-1:
            par.append(map(int,ln.strip().split()))
            ln = f.readline()
        self.fromParents(range(len(par)),par)

    def num_edges(self):
        return sum(map(len,self.edges.values()))

    def p_value(self,g):
        n2 = len(self.vertices)**2
        s1 = self.num_edges()
        s2 = g.num_edges()
        x = 0
        x2 = 0
        for v in self.vertices:
            for w in self.edges[v]:
                if w in g.edges[v]:
                    x+=1
                    try:
                        if self.edge_labelling[(v,w)]==g.edge_labelling[(v,w)]:
                            x2+=1
                    except KeyError:
                        pass
        print s1,s2,x,n2
        print s1,s2,x2,n2
        sens = x*1.0/s2
        spec = 1.0*(n2-s1-s2+x)/(n2-s2)
        return hyp_geom(float(s1),float(s2),float(x),float(n2)),binom_p(x,s1,s2,n2),binom_p(x2,s1,s2,n2,0.33),sens,spec
    
    def weighted_edges(self, weights,fraction=0.0):
        #print "XX",fraction
        res=graph()
        res.vertices=self.vertices
        res.vertice_labelling=self.vertice_labelling
        for u in self.vertices:
            res.edges[u]=[]
        for u in self.vertices:
            if weights: # we have weights for all edges
                s=0.0
                w_dict={}
                if weights[u]:
                    w0,l=weights[u][0]
                    for w,l in weights[u]: #calculate the weights for edges
                        #print w,l
                        w2=2**(w-w0)
                        s+=w2
                        for node in l:
                            try:
                                w_dict[node]+=w2
                            except KeyError:
                                w_dict[node]=w2
                    for v,w in w_dict.items():
                        #print w/s,fraction,v,u
                        if w/s>fraction:
                            res.add_edge(v,u,str(w/s))
        return res
                        
                            
    def to_SIF(self,weights=None):
        result=""
        if weights:
            g1=self.weighted_edges(weights)
        else:
            g1=self
            
        for u in g1.vertices:
            for v in g1.edges[u]:
                label=g1.edge_labelling[(u,v)]
                result+="%s\t%s\t%s\n"%(g1.vertice_labelling[u],label,g1.vertice_labelling[v])
        #print "XX",result,g1.edges
        return result
    
    def from_SIF(self,file):
        self.__init__()
        ln = file.readline()
        while ln:
            rec=ln.strip().split()
            v = self.verticeByLabel(rec[0])
            if v==None: #it's a first time we see vertice v
                v=len(self.vertices)
                self.add_vert(v,rec[0])
            label=rec[1]
            for target in rec[2:]:
                t = self.verticeByLabel(target)
                if t==None: #it's a first time we see vertice t
                    t=len(self.vertices)
                    self.add_vert(t,target)
                self.add_edge(v,t,label)
            ln= file.readline()
        return self

    def labels_2_verts(self):
        new_v=[]
        new_vl = {}
        new_e={}
        new_el={}
        for v in self.vertices:
            l= self.vertice_labelling[v]
            new_v.append(l)
            new_vl[l]=(v,l)
            new_e[l]=[]
            for v2 in self.edges[v]:
                l2 = self.vertice_labelling[v2]
                new_e[l].append(l2)
                new_el[(l,l2)]=self.edge_labelling[(v,v2)]
        g = graph()
        g.vertices=new_v
        g.vertice_labelling=new_vl
        g.edges=new_e
        g.edge_labelling=new_el
        return g

    def compare(self,other):
        #find the intersection of vertices
        g1 = self.labels_2_verts()
        g2 = other.labels_2_verts()
        l = filter(lambda x: x in g2.vertices, g1.vertices)
        g3 = g1.subgraph(l,forward=False)
        g4 = g2.subgraph(l,forward=False)
        #print g3,g4,l
        return  g3.p_value(g4)

    def to_GML(self,f,geom=None,ref=None):
        f.write("graph\t[\n")
        for iv,v in enumerate(self.vertices):
            f.write("\tnode\t[\n")
            f.write("\t\tid\t%d\n"%iv)
            if geom:
                for ln in geom[self.vertice_labelling[v]]:
                    f.write(ln)
            f.write('\t\tlabel\t"%s"\n'%self.vertice_labelling[v])
            f.write('\t]\n')
        for iv,v in enumerate(self.vertices):
            for iu,u in enumerate(self.edges[v]):
                f.write("\tedge\t[\n")
                f.write("\t\ttarget\t%d\n"%self.vertices.index(u))
                f.write("\t\tsource\t%d\n"%iv)
                if not ref:
                    label = ""
                else:
                    if u in ref.edges[v]:
                        label=":V"
                    else:
                        label=":X"
                f.write('\t\tlabel\t"%s"\n'%(self.edge_labelling[(v,u)]+label))
                f.write('\t]\n')
        f.write(']\n')
        
                


def hyp_geom(s1,s2,x,n):
    """
    calculating hypergeometric distribution value using the simplified formula
    """
    def s(s1,s2,i,n):
        res = 1
        for j in range(1,i+1):
            res*=(s1-i+j)*(s2-i+j)/((j)*(n-i+j))
    #        print res
        for j in range(1,s2-i+1):
            res*=(n-s1-s2+i+j)/(n-i-s2+i+j)
    #        print res
        return res
    if s1<s2:
        return hyp_geom(s2,s1,x,n)
    sum = 0
    for i in range(x,s1+1):
        sum +=s(s1,s2,i,n)
    return sum


def fact(n):
    f = 1
    for x in range(n):
        f*=x+1
    return f

def binom(x,n,p=0.5):
     res = 0
     for k in range(x,n+1):
         p_k = p**k*(1-p)**(n-k)
         for i in range(min(k,n-k)):
             p_k*=1.0*(n-i)/(i+1)
             #print n-i,i+1,p_k,k
         res+=p_k
     return res
 
def binom_p(x,s1,s2,N,p=0.5):
    return binom(2*x+N-s1-s2,N,p)

def read_gml_geometry(f):
    def process_node(f):
        ln = f.readline()
        lns=[]
        while ln:
            rec=ln.strip().split()
            if rec[0]=="label":
                label = eval(rec[1])
            if rec[0]=="]":
                return label,lns
            if rec[0]=="graphics":
                while rec[0]!="]":
                    lns.append(ln)
                    ln = f.readline()
                    rec=ln.strip().split()
                lns.append(ln)
            ln = f.readline()
    nodes={}
    ln = f.readline()
    while ln:
        rec=ln.strip().split()
        if rec[0]=="node":
            #next node
            lab,lns=process_node(f)
            nodes[lab]=lns
        ln = f.readline()
    return nodes



# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import graph,fpconst
from heapq import heappush, heappop, heapreplace

class score:
    """
    Abstract class implementing a scoring function.

    To obtain a working implementation, one has to define methods graph_score and data_score
    """
    
    def __init__(self,data_factor = 1.0,prior=None,sloops=0,**kwds):
        self.data_factor=data_factor
        self.sloops = sloops # do we allow self-loops in the networks ?
        
    def graph_score(self,number_of_parents,vertices,weights_of_parents,number_of_data_points, list_of_parents):
        """
        Method for computing graph score factor - ABSTRACT method

        it computes the score given :
        number_of_parents - number of potential parents of a given gene
        vertices - given gene
        par_disc - list of weights_of_parents of parents
        number_of_data_points - number of data points
        """
        pass
    def lower_bound_for_data_score(self,selected_data_empty):
        """
        method for computing the lower bound of the data score factor

        parameters:
        vertices - number of the vertice
        data - list of datapoints :

        each datapoint is a vector containing 2*n_vert values (expression levels for all genes for time points t and t+1
        """
        return 0.0
    
    def data_score(self,selected_data):
        """
        abstract method for computing the data score factor

        parameters:
        gene_vertex - number of the vertice
        numbers_of_parents - list of numbers of parents
        data - list of datapoints :

        each datapoint is a vector containing 2*n_vert values (expression levels for all genes for time points t and t+1
        """
        
        pass


    def subsets(self,list,k):
        """
        generate all the k-long subsets of a list 
        """

        s = graph.stack()
        s.put((list,k,[]))
        try:
            while True: # exception used to get out of the loop
                list,k,acc = s.get()
                if k==0:
                    yield acc
                elif list==[]:
                    pass
                else:
                    s.put((list[1:],k,acc))
                    s.put((list[1:],k-1,acc+[list[0]]))
        except IndexError:
            pass # we've exhausted all options

                    
    def learn_1(self,selected_data,verbose=None,n_min=1,limit=None,score_max=fpconst.PosInf,score_delta=fpconst.PosInf):
        
        if verbose:
            print 'Learning parents of', selected_data.vertex.name, '...',
       
        if not self.sloops:
            selected_data.rm_sloops()
            
        v = selected_data.vertex
        nd = len(selected_data)
        parents=selected_data.parents
        p_weights=selected_data.weights
        n = len(parents)
        try:
            lim=int(limit)
        except TypeError:#limit was None
            lim=n
        
        selected_data_empty=selected_data.subset([])
        mindata = self.lower_bound_for_data_score(selected_data_empty)

        min_set = minset(n_min,score_max,score_delta,self.data_score(selected_data_empty)+\
                self.graph_score(n,v,[],nd,[])) #empty parents set
        if n: # are there any potential parents?
            w_min=p_weights[parents[0]]
            w_max=p_weights[parents[-1]]
            if w_min==w_max:  # we can use algorithm 2
                weight=w_min
                size = 1
                
                mg = self.graph_score(n,v,[weight],nd, [parents[0]])
                while min_set.accepts(mg+mindata) and (size<=lim): #we can possibly add (sub-)optimal scores

                    for sub in self.subsets(parents,size):
                        #print "sub.size ", len(sub)
                        selected_data_sub=selected_data.subset(sub)
                        min_set.add(mg+self.data_score(selected_data_sub), sub)
                    size+=1
                        
                    mg = self.graph_score(n,v,[weight]*size,nd, sub+[parents[0]])
            else: # we have to use algorithm 1
                subsets=[] #successors of considered yet potential parents sets
                for parent in parents: #one-element parents sets
                    #print "one parent"
                    heappush(subsets,(self.graph_score(n,v,[p_weights[parent]],nd, [parent]),[p_weights[parent]],[parent]))
                while subsets:
		    #print subsets
                    mg,weights,sub=heappop(subsets)
		    #print sub
                    if not min_set.accepts(mg+mindata): #we cannot improve the score
                        break
                    selected_data_sub=selected_data.subset(sub)
                    min_set.add(mg+self.data_score(selected_data_sub), sub)
                    #insert sub's successors
                    if len(sub)<lim:
                        last_parent=parents.index(sub[-1])
                        for parent in parents[last_parent+1:]:
                            sub_succ=sub+[parent]
                            weights_succ=weights+[p_weights[parent]]
                            mg_succ=self.graph_score(n,v,weights_succ,nd, sub_succ)
                            heappush(subsets,(mg_succ,weights_succ,sub_succ))
        if verbose:
            print 'done', min_set
        return min_set.optimal, min_set.tolist() 

#    def learn_all(self,vertices,data,n_points):
#        par = {}
#        for v in vertices:
#            par[v] = self.learn_1(v,vertices,data,n_points)
#        return par
    
    def score_graph(self,g,data):
        s = 0.0
        n_vert = len(g.vertices)
        for i,v in enumerate(g.vertices):
            p = g.parents(v)
            selected_data = data.select_1(i,p)
            sg = self.graph_score(n_vert,v,map(lambda par:par.n_disc,p),len(selected_data), p)
            sd = self.data_score(selected_data)
            s+=sg+sd
        return s
        
class minset:
    def __init__(self,size,score_max,score_delta,emptyscore):
        self.optimal=([],emptyscore)
        self.escore=emptyscore
        self.free=size
        self.mscore=score_max
        self.dscore=score_delta
        self.mset=[]
        self.add(emptyscore,[])
        
    def add(self,score,parents):
        if score<min(self.optimal[1]+self.dscore,self.mscore):
            if self.free:
                heappush(self.mset,(-score,parents))
                self.free-=1
                if not self.free:
                    self.mscore=-self.mset[0][0]
            elif self.mset:
                heapreplace(self.mset,(-score,parents))
                self.mscore=-self.mset[0][0]
        if score<self.optimal[1]:
            self.optimal=(parents,score)
            while self.mset and -self.mset[0][0]>self.optimal[1]+self.dscore:
                tmp=heappop(self.mset)
                self.free+=1
                
    def __str__(self):
        self.mset.sort()
        minstr=''
        for (score,parents) in reversed(self.mset):
            minstr+='\n Score '+str(-score)+' for parents set:'
            for p in parents:
                minstr+=' '+p.name
        return minstr
        
    def accepts(self,score):
        return score<min(self.optimal[1]+self.dscore,self.mscore)
        
#    def optimal(self):
#        negscore, parset= max(self.mset)
#        return parset, -negscore
        
    def tolist(self):
        self.mset.sort()
        minlist=[]
        for (score,parents) in self.mset:
            minlist.append((score+self.escore,parents))
        minlist.reverse()
        return minlist
        



# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import stats,math,util

def linspace(minv,maxv,num):
    """
    """
    res=[]
    step=1.0*(maxv-minv)/(num-1)
    v=minv
    while v<=maxv:
        res.append(v)
        v+=step
    return res

class Transform:
    """general (vritual) class for continuous data transformations
    """
    def __init__(self,*args,**kwds):
        """Initialize general parameter of a transformation
        """
        pass
    def estimate(self,vector):
        """Estimate parameters of transformation dependent on input data

        estimates parameters and seves them internally.
        Returns serialized parameters for use with .fromParams method
        """
        pass
    
    def transform(self,vector):
        """Transform a vector of data using internal parameters
        """
        pass
    
    def estTrans(self,vector):
        """Estimate transformation parameters and transform data in one go.

        Only the transformed data are returned. 
        """
        self.estimate(vector)
        return self.transform(vector)
    
    def fromParams(self,params):
        """Fill in internal parameters using serialized data.
        """
        pass
    def getParams(self):
        """Return serialized parameters.
        """
        pass

class BiNormTransform(Transform):
    def __init__(self,t=1.0):
        self.t=t
        
    def transform(self,l):
        trans=lambda x: 1.0/(1.0+(self.p0/self.p1)*math.exp((self.mu1**2-self.mu0**2+2*(self.mu0-self.mu1)*x)/(2*self.t*self.sigma**2)))
        ret=[]
        for x in l:
            try:
                ret.append(trans(x))
            except OverflowError:
                #print "ERROR",x
                ret.append(0.0)
        return ret

    def fromParams(self,p):
        self.mu0,self.mu1,self.sigma,self.p0,self.p1=p
        
    def getParams(self):
        return self.mu0, self.mu1, self.sigma, self.p0, self.p1

    
class BiNormSimple(BiNormTransform):
    def estimate(self,l):
        mu=stats.mean(l)
        self.sigma=stats.stdev(l)/2
        self.mu0=mu-self.sigma
        self.mu1=mu+self.sigma
        self.p0=self.p1=0.5
    
        return self.mu0, self.mu1, self.sigma, self.p0, self.p1

class BiNormMeans(BiNormTransform):
    """Simple estimation of mixture of two gaussians

    Adapted from previous code in 
    """
    def estimate(self,l):
        def sumvar(l,m):
            return stats.sum(map(lambda x:(x-m)**2,l))
        med=stats.lmedianscore(l)
        oldmed=med+1
        while med!=oldmed:
            l0=filter(lambda x: x<med,l)
            l1=filter(lambda x: x>med,l)
            lm=filter(lambda x: x==med,l)
            mu0=stats.mean(l0*2+lm)
            mu1=stats.mean(l1*2+lm)
            (oldmed,med)=(med,(mu0+mu1)/2)
            #print mu0,mu1, oldmed,med#,l0,l1
        self.sigma=max(0.1**10, math.sqrt((sumvar(l0+lm,mu0)+sumvar(l1,mu1)+(mu1-med)**2)/(len(l))))
        self.p0=(len(l0)+len(lm)*0.5)/len(l)
        self.p1=(len(l1)+len(lm)*0.5)/len(l)
        self.mu0=mu0
        self.mu1=mu1
        return self.mu0, self.mu1, self.sigma, self.p0, self.p1


class GaussianMixture(Transform):
    """ Implements a mixture of N Gaussian variables. Estimation done with EM algorithm.

    default N==2, we assume a common variance"""
    def __init__(self,n=2):
        self.n=n
        

    def _Estep(self,mus,vector):
        #assign the nearest mu to each observation
        import collections
        obs=collections.defaultdict(list)
        for v in vector:
            m,val=util.bin_search(v,mus)
            obs[m].append(v)
        return obs
        
    def _Mstep(self,obs):
        #find best self.mus given self.assign
        mus=[]
        for i in range(self.n):
            mus.append(stats.mean(obs[i]))
        return mus
    
    def estimate(self,vector):
        sv=sorted(vector)
        minv=sv[0]
        maxv=sv[-1]
        self.mus=linspace(minv,maxv,self.n+1)
        self.mus=[(x+self.mus[i+1])/2 for i,x in list(enumerate(self.mus))[:-1]]
        self.obs=self._Estep(self.mus,sv)
        mus=self._Mstep(self.obs)
        while mus!=self.mus:
            self.mus=mus #update means
            self.obs=self._Estep(mus,sv) #update observation assignment
            mus=self._Mstep(self.obs) # calculate new mus
        #estimate variance
        self.var=0.0
        for i,mu in enumerate(self.mus):
            for x in self.obs[i]:
                self.var+=(x-mu)**2
        self.var/=len(vector)
        #estimate priors
        self.priors={}
        for x,l in self.obs.items():
            self.priors[x]=len(l)*1.0/len(vector)
        return self.mus,self.var,self.priors
            
    def transform(self,vector):
        """Transformation using multiple gaussian mixture. as of yet unfinished, requires code changes in data.py and score.py
        #TODO: adapt data.py and score.py to allow more general gaussian mixtures
        """

        result=[]
        for v in vector:
            pass
        #trans=lambda x: 1.0/(1.0+(self.p0/self.p1)*math.exp((self.mu1**2-self.mu0**2+2*(self.mu0-self.mu1)*x)/(2*self.t*self.sigma**2)))
        #return map(trans,l)



import random,collections

def combine_fold(folds,i):
    testing=folds[i]
    del folds[i]
    training=reduce(lambda x,y:x+y,folds)
    folds.insert(i,testing)
    return training,testing

def cv_folds(elems,k=10,classes=None,state=None):
    if state!=None: #save (or set) the state of random generator
        random.setstate(state)
    else:
       state=random.getstate()
        
    if classes==None: # if we have a class variable, split list into lists of elements with the same class variable
        lists=[elems]
    else:
        d=collections.defaultdict(list)
        for el,cl in zip(elems,classes):
            d[cl].append(el)
        lists=d.values()

    #shuffle the elements
    shuffled=[]
    for l in lists:
        random.shuffle(l)
        shuffled.extend(l)

    #create folds
    folds=[[] for i in range(k)]
    for i,e in enumerate(shuffled):
        folds[i%k].append(e)    
        
    return state,(combine_fold(folds,i) for i in range(k))
        



# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import math
from score import score

class MDL(score):
    
    def graph_score(self,number_of_potential_parents,gene_vertex,weights_of_parents,number_of_data_points, list_of_parents): # g(Pa)
        graph_structure=(1+len(weights_of_parents))*math.log(number_of_potential_parents+1,2) # = |Pa| * log n
        graph_cpd=0.5*(max(1.1,gene_vertex.n_disc)-1)*math.log(number_of_data_points+1,2) # = 1/2 * log N * (k_x - 1)
        if gene_vertex.cpd in ['or','and','preproc']:
            graph_cpd*=1+sum(weights_of_parents)/2.0
        else:
            for weight in weights_of_parents:
                graph_cpd*=weight
        return graph_structure+graph_cpd
    
    
    def lower_bound_for_data_score(self,selected_data):
        if selected_data.vertex.n_disc:
            return 0.0
        score=0.0
        for d in selected_data.data:
            try:
                score -= d[0]*math.log(d[0],2) + (1-d[0])*math.log(1-d[0],2)
            except:
                pass
        return score*self.data_factor

    def data_score(self,selected_data):
        stats_all,stats_parents = selected_data.stats()
        score = 0.0
        if selected_data.vertex.cpd in ['or','and']: # continuous
            prod_in=(selected_data.vertex.cpd=='or')
            prod_out=1-prod_in
            probabilities=self.cpd_andor(len(selected_data.parents),stats_all,stats_parents,prod_in)
            for a in stats_parents.keys():
                prob_a=probabilities[-1] # last element
                for i,av in enumerate(a):
                    if av==prod_in:
                        prob_a*=probabilities[i]
                try:
                    score-=stats_all[a+(prod_out,)]*math.log(prob_a,2)
                except:
                    pass
                try:
                    score-=stats_all[a+(prod_in,)]*math.log(1-prob_a,2)
                except:
                    pass
        else:
            for a,cv in stats_all.items():
                numbers_of_parents = a[:-1] # except the last one
                cp = stats_parents[numbers_of_parents]
                try:
                    sc = math.log(1.0*cp/cv,2) # - log(cv/cp)
                    score+=cv*sc
                except ZeroDivisionError:
                    pass
        return score*self.data_factor





# Copyright (c) 2004-2009 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#
import math

def safe_exponent(sc):
    """Gives the safe (against overflows) exponent of sc (base 2).

    In case of overflow, give the approximate value.
    """
    try:
        prob=str(2**sc)
    except OverflowError:
        sc10=math.log(2,10)*sc
        prob=str(10**(sc10-int(sc10)))+"e+"+str(int(sc10))
    return prob

def write_indent(file,txt,ind,ln=False):
    """write a string to a file with indent
    """
    file.write(ind)
    file.write(txt)
    if ln:
        file.write("\n")

def writeln_indent(file,txt,ind):
    """write an indented string followed by EOL
    """
    write_indent(file,txt,ind,ln=True)

def pprint(file,obj,level=0,indent="  "):
    """Pretty print a dictionary
    """
    try:
        items=obj.items()
    except AttributeError:
        file.write(repr(obj))
    else:
        writeln_indent(file,"{",level*indent)
        for k,val in sorted(items):
            write_indent(file,repr(k)+" : ",(level+1)*indent)
            pprint(file,val,level+1)
            writeln_indent(file,",","")
        write_indent(file,"}",level*indent)

def bin_search(x,l,b=0,e=None):
    """Finds in a _sorted_ list l the value nearest to x""" 
    if e==None:
        e=len(l)-1
    while b<e-1:
        i=(b+e)/2
        if l[i]<x:
            b=i
        elif l[i]>x:
            e=i
        else: #l[i]==x
            return i,l[i]

    #b==e find the nearer
    if abs(l[b]-x)>abs(l[b+1]-x):
        return b+1,l[b+1]
    else:
        return b,l[b]
        
                               
def rand_exp(n_genes,n_exp,n_disc):
    """Gives random expresssion data.

    n_genes : number of genes
    n_exp : number of experiments
    n_disc : nuimber of discrete levels
    """
    import random
    res = []
    for i in range(n_exp):
        exp = []
        for j in range(n_genes*2):
            exp.append(random.randint(0,n_disc-1))
        res.append(exp)
    return res

def test(n,m,d):
    """Testing function. Old code...
    """
    e = rand_exp(n,m,d)
    return d.learn_1(MDL(),0,range(n),e)



# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import math,fpconst
from score import score
from scipy.stats import chi2

class MIT(score):
    def __init__(self, *args,**kwds):
        score.__init__(self,*args,**kwds)
        self.alpha=kwds["chi_alpha"]
    
    def graph_score(self,number_of_parents,gene_vertex,weights_of_parents,number_of_data_points, list_of_parents): # g(Pa)
        
        # chisquare
        #print list_of_parents
            
        sorted_parents = sorted(list_of_parents, key=lambda gene: gene.n_disc)
        #print "parent.size: ", len(list_of_parents)
        sum_of_l_values = 0.0
        
        for i in range(len(sorted_parents)):            
            l_i_sigma = (gene_vertex.n_disc - 1)*(sorted_parents[0].n_disc - 1)
            for j in range(1,i+1):
                l_i_sigma *= sorted_parents[j].n_disc
            chisquare_value = chi2.isf(1-self.alpha, l_i_sigma)
            #print "chisquare:", chisquare_value
            sum_of_l_values += chisquare_value

        #print [x.name for x in list_of_parents],gene_vertex.name
        #print "g_score: ", sum_of_l_values
        return sum_of_l_values    
            
    def lower_bound_for_data_score(self,selected_data):
        return 0.


    def data_score(self,selected_data):
        stats_all,stats_parents = selected_data.stats()
        stats_for_empty_parents, stats_for_empty_parents_fool = selected_data.subset().stats()
        #print "all",stats_all
        #print "parent",stats_parents
        #print "empty",stats_for_empty_parents

    
   
        score = 0.0
        #print "vertices",[v.name for v in parents_sequence]
        
        if selected_data.vertex.cpd in ['or','and']: # continuous
            raise Exception
        else:
            number_of_effective_observations = len(selected_data)
            #compute myEntropy
            H=0.0
            for k,v in stats_for_empty_parents.items():
                p=float(v)/number_of_effective_observations
                H-=p*math.log(p,2)
            #print H
            if len(selected_data.parents)==0: # CMI(X,[])==0
                score=H
            else:
                #compute CMI
                CMI=0.
                for obs_values,count in stats_all.items():
                    obs_parents=obs_values[:-1]
                    obs_vert=obs_values[-1]
                    p_xy=1.*count/number_of_effective_observations
                    p_x=1.*stats_parents[obs_parents]/number_of_effective_observations
                    p_y=1.*stats_for_empty_parents[(obs_vert,)]/number_of_effective_observations
                    CMI+=p_xy*math.log(p_xy/(p_x*p_y),2)
                    #print obs_parents,obs_vert,p_xy,p_x,p_y,p_xy*(math.log(p_xy/(p_x*p_y))),CMI
                score=H-CMI
            
        return 2*number_of_effective_observations*score*self.data_factor
        #print "d_score: ", score
        #return score





import rpy2.robjects as robjects

robjects.r("library(ROCR)")
robjects.r('auc<-function(p,s){performance(p,"auc",fpr.stop=s)}')

def device(filename):
    robjects.r.pdf(file=filename,width=5,height=5)

def close():
    robjects.r("dev.off()")

def aucROC(predictions,labels,part=1.):
    try:
        l= robjects.IntVector(labels)
    except:
        print "Problem with data",labels
    else:
        try:
            p= robjects.FloatVector(predictions)
        except:
            print "Problem with data",predictions
        else:
            pred = robjects.r.prediction(p,l)
            
            return robjects.r.auc(pred,part).do_slot("y.values")[0][0]

def simpleROC(predictions,labels,title,y="tpr",x="fpr",diag=True):
    try:
        l= robjects.IntVector(labels)
    except:
        print "Problem with data",labels
    else:
        try:
            p= robjects.FloatVector(predictions)
        except:
            print "Problem with data",predictions
        else:
            pred = robjects.r.prediction(p,l)
            perf = robjects.r.performance(pred,y,x)
            auc  = robjects.r.performance(pred,"auc").do_slot("y.values")[0][0]
            print "AUC",auc
            title="%s (AUC=%f)"%(title,auc)
            try:
                robjects.r.plot(perf,main=title,colorize=True)
            except:
                print "problem with plotting"
            if diag:
                robjects.r.abline(a=0,b=1)
    
    
def cvROC(predictions,labels,title,y="tpr",x="fpr",diag=True):
    #remove folds with only positive or only negative labels
    bad_folds=[]
    for i in range(len(labels)):
        if (True in labels[i]) and (False in labels[i]):
            pass
        else:
            bad_folds.append(i)
    print bad_folds
    for i in reversed(bad_folds): #remove in reversed order not to break the indexing
        del predictions[i]
        del labels[i]
        
    #Now do the ROC curve
    try:
        rlabs=map(robjects.IntVector,labels) #make IntVector out of each CV fold labels
        ll=robjects.r.list(*rlabs)
    except:
        print "Problem with data",labels
    else:
        try:
            rpreds=map(robjects.FloatVector,predictions)
            pp=robjects.r.list(*rpreds)
        except:
            print "Problem with data",predictions
        else:
            
            try:
                pred = robjects.r.prediction(pp,ll)
                perf = robjects.r.performance(pred,y,x)
            except:
                print "problem with ROCR",predictions,labels
            else:
                try:
                    robjects.r.plot(perf,main=title,col="grey82")
                    robjects.r.plot(perf,lwd=3,avg="vertical",add=True)
                    if diag:
                        robjects.r.abline(a=0,b=1)
                except:
                    print "problem with plotting"


def top_n(predictions,labels,title):
    try:
        l= robjects.IntVector(labels)
    except:
        print "Problem with data",labels
    else:
        try:
            p= robjects.FloatVector(predictions)
        except:
            print "Problem with data",predictions
        else:
            pred = robjects.r.prediction(p,l)
            perf = robjects.r.performance(pred,"prec","rec")
            xvs=perf.do_slot("x.values")[0]
            yvs=perf.do_slot("y.values")[0]
            print "AUC",auc
            title="%s (AUC=%f)"%(title,auc)
            try:
                robjects.r.plot(perf,main=title,colorize=True)
            except:
                print "problem with plotting"
            robjects.r.abline(a=0,b=1)



#!/usr/bin/env python2.5
# Copyright (c) 2004-2007 Bartek Wilczynski and Norbert Dojer; All Rights Reserved.
#
# This software is distributable under the terms of the GNU
# General Public License (GPL) v2, the text of which can be found at
# http://www.gnu.org/copyleft/gpl-2.0.txt. Installing, importing or otherwise
# using this module constitutes acceptance of the terms of this License.
#
# Disclaimer
# 
# This software is provided "as-is".  There are no expressed or implied
# warranties of any kind, including, but not limited to, the warranties
# of merchantability and fittness for a given application.  In no event
# shall the authors be liable for any direct, indirect, incidental,
# special, exemplary or consequential damages (including, but not limited
# to, loss of use, data or profits, or business interruption) however
# caused and on any theory of liability, whether in contract, strict
# liability or tort (including negligence or otherwise) arising in any way
# out of the use of this software, even if advised of the possibility of
# such damage.
#

import graph,stats,math,fpconst
from itertools import chain
from BDE import BDE
from MDL import MDL
from MIT import MIT
import continuous
import util
from math import log


def eval_func(f_args):
    """Takes a tuple of a function and args, evaluates and returns result"""
    return f_args[0](*f_args[1:])

def learn_x(v):
    x,self,min_empty,min_optim,scr,verbose,n_min,limit,score=v

    selected_data = self.select_1(x)

    if min_empty:
        score_empty=scr.data_score(selected_data.subset([]))+scr.graph_score(len(parents),x,[],len(selected_data))
        score_max=score_empty-math.log(min_empty,2)
    else:
        score_max=fpconst.PosInf
    if min_optim:
        score_delta=-math.log(min_optim,2)
    else:
        score_delta=fpconst.PosInf
    (par,sc),minlist = scr.learn_1(selected_data,verbose,n_min,limit,score_max,score_delta)
    return x,par,minlist,score 



class experiment:
    """Class representing an experiment, possibly containing more than one time-point
    """
    def __init__(self,name,data,perturbed,points=[],static=False):
        """Constructor

        retains the argument data as self._data, while filling the self.data with double vectors
        (duplicates for static networks, consecutive pairs for dynamic)
        """
        self._data=data
        self.perturbed=perturbed
        self.data=[]
        self.name=name
        self.points=points
        if static:
            for i,d in enumerate(self._data):
                self.data.append(self._data[i]+d)
        else:
            # in format (data from time t - 1, data from time t)
            for i,d in enumerate(self._data[1:]):
                self.data.append(self._data[i]+d)


class dataset_1:
    """Class representing data selected for learning parents of one vertex
    """
    def __init__(self,data,vertex,parents,weights=None):
        """Constructor
        """
        self.data=data
        self.vertex=vertex
        self.parents=parents
        self.weights=weights
        
    def __len__(self):
        return len(self.data[-1])
        
    def orientation(self):
        res={}
        for i,p in enumerate(self.parents):
            sign = stats.pearsonr(self.data[i],self.data[-1])[0]
            if sign >=0:
                res[p]="+"
            else:
                res[p]="-"
        return res
        
    def subset(self,sub=[]):
        ssub = set(sub)
        sub_ind=[i for i,p in enumerate(self.parents) if p in ssub]
        ordered_sub=[self.parents[i] for i in sub_ind]
        data = [self.data[i] for i in sub_ind]
        data.append(self.data[-1])
        return dataset_1(data,self.vertex,ordered_sub,self.weights)
        
    def rm_sloops(self):
        try:
            i=self.parents.index(self.vertex)
            del self.parents[i]
            del self.data[i]
        except ValueError:
            pass
        
    def stats(self):
        def strec(key,prob,d):
            first,first_disc=d[0]
            rest=d[1:]
            if rest==[]:
                try:
                    stats_parents[key]+=prob
                except KeyError:
                    stats_parents[key]=prob
                if first_disc:
                    try:
                        stats_all[key+(first,)]+=prob
                    except KeyError:
                        stats_all[key+(first,)]=prob
                else:
                    try:
                        stats_all[key+(0,)]+=prob*(1-first)
                    except KeyError:
                        stats_all[key+(0,)]=prob*(1-first)
                    try:
                        stats_all[key+(1,)]+=prob*first
                    except KeyError:
                        stats_all[key+(1,)]=prob*first
            else:
                if first_disc:
                    strec(key+(first,),prob,rest)
                else:
                    strec(key+(0,),prob*(1-first),rest)
                    strec(key+(1,),prob*first,rest)
        d_disc=map(lambda p: p.n_disc,self.parents+[self.vertex])
        if 0 not in d_disc:
            return self.stats_disc()
        stats_all = {}
        stats_parents = {}
        for d in zip(*self.data):
            strec((),1,zip(d,d_disc))
        return stats_all,stats_parents
        
    def stats_disc(self):
        stats_all = {}
        stats_par = {}
        for d in zip(*self.data):
            d_par = tuple(d[:-1])
            d_all = tuple(d)
            try:
                stats_par[d_par]+=1
            except KeyError:
                stats_par[d_par]=1
            try:
                stats_all[d_all]+=1
            except KeyError:
                stats_all[d_all]=1
        return stats_all,stats_par
  

class gene:
    """Class corresponding to a single variable

    #TODO: rename to variable or smth... gene is inappropriate
    """
    def __init__(self,name,n_disc,index):
        self.name=name
        self.n_disc=n_disc
#        self.parents=[]
        self.values=[]
        self.index=index
        self.cpd=None
        
    def __str__(self):
        return self.name
    
    def disc_vals(self):
        """Returns discretized values of a variable

        works also for continuous variables
        """
        if self.values:
            try:
                return map(int,self.values)
            except ValueError:
                return self.values
        else:
            return [0,1] #This should take care of continuous variables... may need a change if we introduce more complex mixtures.

    def __index__(self):
        """returns the index of a variable.
        """
        return self.index

    def __add__(self,other):
        """Implements addition on indexes, so that we can write table[var+1] instad of table[var.index+1]
        
        """
        return other+self.index
    def __hash__(self):
        """returns the hash value for comparisons
        """
        return self.index

    def __eq__(self,other):
        if type(other)==int:
            return self.index==other
        else:
            return self.index==other.index    
        
class dataset:
    """
    A class representing a complete dataset.

    Implements input, output, data selection and also functions for learning networks given a scoring function
    """
    def __init__(self,name='no_name'):
        self.name=name
        self.n_series=0
        self.n_gen=0
        self.points=[]
        self.vertices= []
        self.vertice_names=[]
        self.static=False
        self.cont_weight=1.5
        self.regulators=[]
        self.parents={}
        self.prior={}
        
    def fromFile(self,f,float_transform=continuous.BiNormMeans()):
        import warnings
        warnings.warn("The old data format is deprecated! Please use the new format.")
        #number of genes in the first line
        self.n_gen=int(f.readline())
        # experiments description
        ln = f.readline().strip().split()
        self.n_series = int(ln[0])
        series = []
        for txt in ln[1:]:
            if txt.find("[")==-1:
                series.append((int(txt),[]))
            else: #perturbed genes
                series.append((int(txt[:txt.find("[")]),
                               [int(txt[1+txt.find("["):txt.find("]")])]))
        lines = []
        self.vertices=[]
        for i in range(self.n_gen):
            ln = f.readline().strip().split()
            n_disc = max(0,int(ln[1]))
            self.vertices.append(gene(ln[0],n_disc,i))
            if n_disc==0:
                if float_transform: #use float_transformn if specified
                    lines.append(float_transform.estTrans(map(float,ln[2:])))
                else:
                    lines.append(map(float,ln[2:]))
            else:
                lines.append(map(int,ln[2:]))
        self.points=[]
        i=0
        for s,p in series:
            d = []
            for k in range(s): 
                d.append(map(lambda x:x[i],lines))
                i+=1
            name="serie%d"%len(self.points)
            self.points.append(experiment(name,d,p,range(len(d))))

#        for v in self.vertices:
#            v.parents=map(lambda v: (float(max(self.cont_weight,v.n_disc)),v),self.vertices)
        for e in self.points:
            e.perturbed=map(lambda i: self.vertices[i],e.perturbed)
        return self

    def toNewFile(self,file):
        #perturbations
        for s in self.points:
            for g in s.perturbed:
                file.write("#perturbed\t%s\t%s\n"%(s.name,self.vertices[g]))
        #condition names\
        file.write("conditions")
        for s in self.points:
            for i,d in enumerate(s._data):
                file.write("\t%s:%d"%(s.name,i))
        file.write("\n")
        #expr_data
        for i,v in enumerate(self.vertices):
            file.write(str(v))
            for s in self.points:
                for j,d in enumerate(s._data):
                    if v.n_disc>0: #discrete
                        format="\t%d"
                    else: #continuous
                        format="\t%f"
                    file.write(format%d[i])
            file.write("\n")

    def fromNewFile(self,file,float_transform=continuous.BiNormMeans()):
        pert=[]
        disc={}
        cpd={'and':[],'or':[]}
        contin=[]
        self.default=None
        rec=[]
        regul=[]
        while not rec:
            ln=file.readline()
            rec=ln.strip().split()
        #################
        #THE PREAMBLE
        ################
        #print 'preamble'
        while rec[0][0]=="#": #preamble
            if rec[0]=="#perturbed": #experiment serie name followed by list of its perturbed vertices
                pert.append((rec[1],rec[2:]))
            elif rec[0]=="#regulators": #list of potential regulators of all vertices (except specified in #parents command) not specified in previous or present #regulators command
                reg=filter(lambda x:True not in map(lambda r:x in r,regul),rec[1:])
                for i,r in enumerate(reg):
                    if i!=reg.index(r):
                        reg[i]=None
                reg=filter(None,reg)
                regul.append(reg)
            elif rec[0]=="#parents": #vertice name followed by list of its potential regulators
                reg=rec[2:]
                for i in range(len(reg)):
                    if i!=reg.index(reg[i]):
                        reg[i]=None
                reg=filter(None,reg)
                self.parents[rec[1]]=reg
            elif rec[0]=="#default": #list of default possible values or 'FLOAT' for default continuous values
                if len(rec)==2 and rec[1]=="FLOAT":
                    self.default=[]
                else:
                    self.default=rec[1:]
            elif rec[0]=="#discrete": #vertice name followed by list of all its possible values
                disc[rec[1]]=rec[2:]
            elif rec[0]=="#continuous": #list of vertices with continuous values
                contin.extend(rec[1:])
            elif rec[0]=="#and": #list of vertices with cpd noisy-and
                cpd['and'].extend(rec[1:])
            elif rec[0]=="#or": #list of vertices with cpd noisy-or
                cpd['or'].extend(rec[1:])
            elif rec[0]=="#priorvert": #prior weight of all outgoing edges (except specified in #prioredge command) followed by list of related vertices
                val=float(rec[1])
                for name in rec[2:]:
                    self.prior[name]=val
            elif rec[0]=="#prioredge": #prior weight of edges preceded by edge target (common for all specified edges) and followed by list of edge sources
                target_name=rec[1]
                val=float(rec[2])
                for name in rec[3:]:
                    self.prior[(target_name,name)]=val
            else:
                pass #ignoring other stuff
            rec=[]
            while not rec:
                ln=file.readline()
                rec=ln.strip().split()
        self.regulators=regul
        
        ################
        #condition names
        ################
        #print 'conditions'
        if rec[0]!='conditions':
            self.name=rec[0]
        #dynamic or static data?
        if len(rec[1].split(":"))==1:
            self.static=True
        #else:
        #    self.static=False
        if self.static:
            conds=map(lambda x : [x],rec[1:])
            names=conds
        else:
            # todo:pbd serie not used any more?
            serie=rec[1].split(":")[0]
            # pbd: explain why conds is a singleton and not just a single standalone element
            cond=[rec[1].split(":")[1]]
            conds=[cond]
            names=[serie]
            for c in rec[2:]:
                r=c.split(":")
                if r[0]==serie:
                    cond.append(r[1])
                else:
                    serie=r[0]
                    names.append(serie)
                    cond=[r[1]]
                    conds.append(cond)
            # pbd: after that conds looks like [['1', '2', '3', ...]]
        
        ln=file.readline()
        ###################
        # EXPRESSION DATA #
        ###################
        #print 'expression'
        self.vertices=[]
        self.vertice_names=[]
        #prepare the lists for expression data
        n_cond=sum(map(len,conds))
        exp_data=map(lambda l: [ [] for el in l],conds)
        geneindex=0
        while ln: #expression data
            rec=ln.strip().split()
            if not rec:
                ln=file.readline()
                continue
            name=rec[0]
            vals=[]
            #are the values discrete or not?
            if name in contin:#did we say it's continuous?
                discrete=False
            elif name in disc.keys(): #did we say is's discrete?
                discrete=True
                vals=disc[name]
            elif self.default!=None:# did we specify a default
                if self.default==[]:
                    discrete=False
                else:
                    discrete=True
                    vals=self.default
            else: #We did not specify the type of values for this gene, try to guess it
                try:
                    i = int(rec[1])
                except ValueError: #it's not an integer
                    try:
                        i=float(rec[1])
                    except ValueError: #it's not a float -> it's an alphanumerical value
                        discrete=True
                    else:
                        discrete=False # it's a float
                else:
                    discrete=True #it's an int, let's map the strings to ints
            if discrete:
                line=rec[1:]
                if vals==[]:#we don't know what are the possible values
                    for v in line:
                        if v not in vals:
                            vals.append(v)
                    #let's sort the values
                    vals.sort()
            else: #not discrete
                line=map(float,rec[1:])
                float_params=[]
                if float_transform: #use float_transform if specified
                    float_params=float_transform.estimate(line)
                    line=float_transform.transform(line)
                
            self.vertices.append(gene(name,len(vals),geneindex))
            self.vertices[-1].values=vals
            if discrete==False:
                self.vertices[-1].floatParams=float_params
            else:
                self.vertices[-1].floatParams=None
            self.vertice_names.append(name)
            #append this line to the expression data
            for el,l in zip(line,chain(*exp_data)):
                if discrete:
                    l.append(vals.index(el))
                else:
                    l.append(el)
            ln=file.readline()
            geneindex+=1
        ################
        #POSTPROCESSING#
        ################
        #print 'postprocessing'
        #process regulatory constraints
        for v in self.vertices:
            #specify cpd
            if v.name in cpd['and']:
                v.cpd='and'
            elif v.name in cpd['or']:
                v.cpd='or'
        #process the lists of perturbed genes and make the list of experiments
        for i,(c,n) in enumerate(zip(conds,names)):
            pg = []
            for x,p in pert:
                if x==n:
                    pg.extend(map(lambda n: self.vertices[self.vertice_names.index(n)],p))# map gene names to indexes
            
            self.points.append(experiment(n,exp_data[i],pg,c,self.static))

        #set the number of genes
        self.n_gen=geneindex
        return self

    def select_1(self,vertex,parents=None):
        """
        given a vertice and a list of his parents,
        select parents values at time t
        and child value at time t+1,
        omitting the experiments where the child was perturbed.
        """
        if parents is None:
            parents = self.get_potential_parents(vertex)
        #assign weights to parents
        weights={}
        for par in parents:
            weight=float(max(self.cont_weight,par.n_disc))
            if (vertex.name,par.name) in self.prior.keys():
                weight=weight**self.prior[(vertex.name,par.name)]
            elif par.name in self.prior.keys():
                weight=weight**self.prior[par.name]
            weights[par]=weight
        parents.sort(key = lambda p: weights[p])

        res = []
        for expment in self.points:
            if vertex not in expment.perturbed:
                for sample in expment.data:
                    res.append([sample[par] for par in parents]+[sample[vertex+self.n_gen]])
        return dataset_1(zip(*res),vertex,parents,weights)

    # pbd:debug there are some inconsistencies between this version of getting potential parents and
    # the one implemented in bnfinder-waszczuk branch - see how it works on input3.txt or input4.txt 
    def get_potential_parents(self, node):
        """selects potential parents for node"""
        if self.parents.has_key(node.name):
            parents=map(lambda n: self.vertices[self.vertice_names.index(n)],self.parents[node.name])
        else:
            #maybe there are regulators specified
            if self.regulators==[]:
                parents=[v for v in self.vertices]
            else:#there is a list of regulators
                parents=[]
                for r in self.regulators:
                    if node.name in r:
                        break
                    else:
                        parents.extend(map(lambda n: self.vertices[self.vertice_names.index(n)],r))
        return parents

 
    def learn(self,score,data_factor,prior=None,distrib=False,\
            verbose=None,n_min=1,limit=None,min_empty=None,min_optim=None,fdr_weights=None):
        
        scr=score
        
        if distrib:
            try:
                from multiprocessing import Pool#,Queue
            except ImportError:
                print "Problem invoking multiprocessing module. Running on a single CPU"
                distrib=False
                #from Queue import Queue
            else:
                pool=Pool(distrib)
                mapf=pool.map


        score = 0.0
        pars ={}
        subpars ={}


        # fdr weights have higher priority than self.prior
        # if both set - ignore later one
        if self.prior and fdr_weights:
            print "Warning: fdr parameter set - ignoring priors in dataset"

        if fdr_weights:
            self.prior = {}
            for k in fdr_weights:
                self.prior[k] = log(fdr_weights[k])

        if distrib:
            from itertools import izip,repeat
            result=pool.map(learn_x,[(x,self,min_empty,min_optim,scr,verbose,n_min,limit,score) for x in self.vertices])
                            #eval_func,izip([learn_x]*len(self.vertices),self.vertices))
            #pool.close() #raises weird errors...., not necessary in normal bnf operation, since the process exits after learning
        else:
            result=map(learn_x,[(x,self,min_empty,min_optim,scr,verbose,n_min,limit,score) for x in self.vertices])

        #print result
        self.vertices=[r[0] for r in result]
        
        for x,par,minlist,score in result:
            pars[x]=par
            subpars[x]=minlist
            score+=score
        #print pars
        #move parents from a dictionary to a list
        par_list = []
        for v in self.vertices:
            try:
                par_list.append(pars[v.__index__()])
            except KeyError:
                print "ERRR",par_list,v.name,id(v),v.__index__(),[map(lambda x:(x.name,id(x)),l) for l in pars.values()]
        pars=par_list

        g = graph.graph()
        g.fromParents(self.vertices ,pars)
        for v,par in zip(self.vertices,pars):
            g.vertice_labelling[v]=v.name
            v_orientation=self.select_1(v,par).orientation()
            for p in par:
                g.edge_labelling[p,v]=v_orientation[p]
        return score,g,subpars
   
    def get_stats(self,g):
        stats={}
        for v in self.vertices:
            stats[v] = self.select_1(v,g.parents(v)).stats()
        return stats
        
    def write_txt(self,subpars,file_name):
        """Outputs suboptimal parents sets into a text file
        """

        f=open(file_name,"w")
        for v,minlist in subpars.items():
            f.write('\n%s'% v.name)
            for prob, pars in minlist:
                prob_s=util.safe_exponent(prob)
                f.write('\n %s '% prob_s)
                for p in pars:
                    f.write(' %s'% p.name)
        f.close()

    def cpd_andor(self,n_par,stats_all,stats_par,prod_in,eps=10**(-7)):
        n_all=n_par+1
        epsilon=eps*n_all*10
        v_all=0.0
        v_in=0.0
        for all,val in stats_all.items():
            v_all+=val
            if all[-1]==prod_in:
                v_in+=val
        if v_all-v_in<=eps:
            return [0.0]*n_all
        elif v_in<=eps:
            return [1.0]*n_all
        p=[1.0-eps]*n_all
        delta=epsilon+1.0
        while delta>epsilon:
            delta=0.0
            for i in range(n_all):
                p_old=p[i]
                delt=eps+1.0
                while abs(delt)>eps:
                    sumf=0.0
                    sumfprim=0.0
                    for par in stats_par.keys():
                        all=par+(prod_in,)
                        if all[i]==prod_in:
                            if all in stats_all.keys():
                                prod=1.0
                                for j in range(n_all):
                                    if all[j]==prod_in:
                                        prod*=p[j]
                                q=stats_all[all]/(1.0-prod)
                                sumf+=q-stats_par[par]
                                sumfprim+=(q/(1.0-prod))*(prod/p[i])
                            else:
                                sumf-=stats_par[par]
                    delt=sumf/sumfprim
                    p[i]-=delt
                    if p[i]>=1.0:
                        delt=1.0-eps-p[i]-delt
                        p[i]=1.0-eps
                delta+=abs(p[i]-p_old)
        return p

    def to_cpd(self,g,dirichlet=None):
        if dirichlet==None:
            dirichlet=1.0
        else:
            dirichlet=float(dirichlet)
        result={}
        for v in self.vertices:
            if True:#if v.name not in self.regulators[0]:
                n_disc=max(2,v.n_disc)
                stats_all,stats_par = self.select_1(v,g.parents(v)).stats()
                var_d={}
                result[str(v)]=var_d
                var_d["vals"]=v.values
                var_d["pars"]=map(str,g.parents(v))
                var_d["floatParams"]=str(v.floatParams)
                if v.cpd=='and':
                    var_d["type"]="and"
                elif v.cpd=='or':
                    var_d["type"]="or"
                var_cpds={}
                var_d["cpds"]=var_cpds
                if v.cpd:
                    p=self.cpd_andor(len(g.parents(v)),stats_all,stats_par,v.cpd=='or')
                    for (i,pv) in enumerate(p[:-1]):
                        var_cpds[str(g.parents(v)[i])]=pv
                    var_cpds[None]=p[-1]
                else:
                    stats_values={}
                    for par_val in stats_par.keys():
                        stats_values[par_val]=[]
                    for all_val in stats_all.keys():
                        stats_values[all_val[:-1]].append(all_val[-1])
                    for par_val in stats_par.keys():
                        d={}
                        var_cpds[par_val]=d
                        for val in stats_values[par_val]:
                            d[val]=(stats_all[par_val+(val,)]+dirichlet)/(stats_par[par_val]+dirichlet*n_disc)
                        d[None]=dirichlet/(stats_par[par_val]+dirichlet*n_disc)
                    var_cpds[None]= 1.0/n_disc
        return result

        
        
    def write_cpd(self,g,f,dirichlet=None):
        if dirichlet==None:
            dirichlet=1.0
        else:
            dirichlet=float(dirichlet)
        f.write('{\n')
        for v in self.vertices:
            n_disc=max(2,v.n_disc)
            stats_all,stats_par = self.select_1(v,g.parents(v)).stats()
            f.write('\''+str(v)+'\' : {\n')
            f.write('  \'vals\' : '+str(v.values)+' ,\n')
            f.write('  \'pars\' : '+str(map(str,g.parents(v)))+' ,\n')
            if v.cpd=='and':
                f.write('  \'type\' : \'and\' ,\n')
            elif v.cpd=='or':
                f.write('  \'type\' : \'or\' ,\n')
            f.write('  \'cpds\' : {\n')
            if v.cpd:
                p=self.cpd_andor(len(g.parents(v)),stats_all,stats_par,v.cpd=='or')
                for (i,pv) in enumerate(p[:-1]):
                    f.write('     '+str(g.parents(v)[i])+' : '+str(pv)+'\n')
                f.write('     '+str(v)+' : '+str(p[-1])+' } } ,\n')
            else:
                stats_values={}
                for par_val in stats_par.keys():
                    stats_values[par_val]=[]
                for all_val in stats_all.keys():
                    stats_values[all_val[:-1]].append(all_val[-1])
                for par_val in stats_par.keys():
                    f.write('     '+str(par_val)+' : { ')
                    for val in stats_values[par_val]:
                        f.write(str(val)+' : '+str((stats_all[par_val+(val,)]+dirichlet)/(stats_par[par_val]+dirichlet*n_disc))+' , ')
                    f.write('None : '+str(dirichlet/(stats_par[par_val]+dirichlet*n_disc))+' } ,\n')
                f.write('     None : '+str(1.0/n_disc)+' } } ,\n')
        f.write('}\n')


    def write_bif(self,g,file_name,dirichlet=None,comments=[]):
        if dirichlet==None:
            dirichlet=1.0
        else:
            dirichlet=float(dirichlet)
        f=open(file_name,"w")
        f.write('\\\\ File generated by BNfinder\n')
        import time
        f.write('\\\\ '+time.strftime('%x %X')+'\n')
        for com in comments:
            f.write('\\\\ '+com+'\n')
        f.write('\\\\ Conditional probability distributions generated with total pseudocounts number %f\n' % dirichlet)
        f.write('\n')
        f.write('network \"%s\" {}\n' % self.name)
        f.write('\n')
        for v in self.vertices:
            n_disc=max(2,v.n_disc)
            stats_all,stats_par = self.select_1(v,g.parents(v)).stats()
            f.write('variable \"%s\" {\n' % str(v))
            f.write('   type discrete[%d] { \"%s\" }\n'% (n_disc,'\" \"'.join(map(str,v.disc_vals()))))
            f.write('    }\n')
            if g.parents(v):
                f.write('probability ( \"%s\" | \"%s\" ) {\n'% (str(v),'\" \"'.join(map(str,g.parents(v)))))
                f.write('     default %s ;\n' % ' '.join([str(1.0/n_disc)]*n_disc))
            else:
                f.write('probability ( \"%s\" ) {\n'% str(v))
            stats_values={}
            for par_val in stats_par.keys():
                stats_values[par_val]=[]
            for all_val in stats_all.keys():
                stats_values[all_val[:-1]].append(all_val[-1])
            for par_val in stats_par.keys():
                if par_val:
                    f.write('     ( \"%s\" ) ' % '\" \"'.join(map(str,par_val)))
                else:
                    f.write('     table ')
                for val in v.disc_vals():
                    if val in stats_values[par_val]:
                        f.write(str((stats_all[par_val+(val,)]+dirichlet)/(stats_par[par_val]+dirichlet*n_disc))+' ')
                    else:
                        f.write(str(dirichlet/(stats_par[par_val]+dirichlet*n_disc))+' ')
                f.write(';\n')
            f.write('    }\n\n')
        f.close()

    def point2dict(self,i):
        """returns a dictionary representing one (i-th) data-point
        """
        return dict(zip(self.vertice_names+map(lambda x: x+"'",self.vertice_names),self.points[i].data[0]))
    def get_vertice_by_name(self,name):
        """Returns vertex of self corresponding to the given name
        """
        return self.vertices[self.vertice_names.index(name)]
    
    def get_probs(self,d,cpd,pars):
        """returns list of posterior probabilities of all parents being in all possible states.

        for discrete variables gives singular distributions, important for continuous vars.
        """
        res=[]
        import collections
        for p in pars:
            v=self.get_vertice_by_name(p)
            if str(d[p]) in v.values:
                dd=collections.defaultdict(float)
                dd[d[p]]=1.0
                res.append(dd) #singular distribution for discrete vars
            else: #continuous vars need care
                model=continuous.BiNormMeans()
                model.fromParams(eval(cpd[p]["floatParams"]))
                x=model.transform([d[p]])[0] # x -> probability of parent being 1
                res.append({1:x,0:1-x})
#        print res
#        a=b
        return res
                
            
    def classify(self, cpd,f,ML=False,prob=None):
        """
        Output the classification based on the regulators' values from dataset and probabilities from a cpd.

        if ML==True:
           output only the most likely value
        else: (Default)
           output posterior probabilities
        if prob:
           output only the probability of a sigle class
        """
        f.write("classes")
        for p in self.points:
            f.write("\t%s"%p.name[0])
        f.write("\n")
        dicts=map(self.point2dict,range(len(self.points)))
        for v in cpd.keys():
            if v not in self.vertice_names:
                f.write("%s\t"%v)
                c=cpd[v]
                for i,p in enumerate(self.points):
                    total_prob={}
                    for val in c["vals"]:
                        total_prob[int(val)]=0.0
                    probs=self.get_probs(dicts[i],cpd,c["pars"])
                    for possib,poster in c["cpds"].items(): #possible parent values
                        if possib!=None:
                            r=reduce(lambda x,y:x*y,[d[x] for (x,d) in zip(possib,probs)],1)
                            for val in total_prob.keys():
                                try:
                                    total_prob[val]+=r*poster[val]
                                except KeyError:
                                    pass
                                
                    ret_d=total_prob
                    #print ret_d
                    if ML: # Max likelihod
                        f.write("%s\t"%(str(sorted(zip(*reversed(zip(*(ret_d.items())))))[-1][-1]))) # pick the arg max from ret_d
                    elif prob!=None: # the probability of a given value
                        try:
                            f.write("%s\t"%(str(ret_d[prob])))
                        except KeyError:
                            f.write("%s\t"%"-0.0")
                    else: # output whole posterior
                        f.write("%s\t"%(str(ret_d)))
                    #par_vals=tuple(map(lambda x: dicts[i][x],c["pars"]))
                    #try:
                    #    ret_d=c["cpds"][par_vals]
                    #except KeyError:
                        #testing, to be removed
                        #print"*",
                        #model=continuous.BiNormMeans()
                        #print cpd[c["pars"][0]]["floatParams"]
                        #model.fromParams(eval(cpd[c["pars"][0]]["floatParams"]))
                        #x=model.transform(par_vals)[0] # x -> probability of parent being 1
                        #p=x*c["cpds"][1,][1]+(1-x)*c["cpds"][0,][1]
                        #f.write("%f\t"%(p))
                    #    f.write("%s\t"%"0.0")
                    #else:
#                         if ML: # Max likelihod
#                             f.write("%s\t"%(str(sorted(zip(*reversed(zip(*(ret_d.items())))))[-1][-1]))) # pick the arg max from ret_d
#                         elif prob!=None: # the probability of a given value
#                             try:
#                                 f.write("%s\t"%(str(ret_d[prob])))
#                             except KeyError:
#                                 f.write("%s\t"%"0.0")
#                         else: # output whole posterior
#                             f.write("%s\t"%(str(ret_d)))
                f.write("\n")
    def classify_list(self, cpd,ML=False,prob=None):
        """
        Output the classification based on the regulators' values from dataset and probabilities from a cpd.

        if ML==True:
           output only the most likely value
        else: (Default)
           output posterior probabilities
        if prob:
           output only the probability of a sigle class
        """
        result={}
        dicts=map(self.point2dict,range(len(self.points)))
        for v in cpd.keys():
            if v not in self.vertice_names:
                c=cpd[v]
                for i,p in enumerate(self.points):
                    total_prob={}
                    for val in c["vals"]:
                        total_prob[int(val)]=0.0
                    probs=self.get_probs(dicts[i],cpd,c["pars"])
                    for possib,poster in c["cpds"].items(): #possible parent values
                        if possib!=None:
                            r=reduce(lambda x,y:x*y,[d[x] for (x,d) in zip(possib,probs)],1)
                            for val in total_prob.keys():
                                try:
                                    total_prob[val]+=r*poster[val]
                                except KeyError:
                                    pass
                                
                    ret_d=total_prob
                    #print p.name[0],v
                    if ML: # Max likelihod
                        result[p.name[0],v]=sorted(zip(*reversed(zip(*(ret_d.items())))))[-1][-1] # pick the arg max from ret_d
                    elif prob!=None: # the probability of a given value
                        try:
                            result[p.name[0],v]=(ret_d[prob])
                        except KeyError:
                            result[p.name[0],v]=0.0
                    else: # output whole posterior
                        result[p.name[0],v]=ret_d
        return result
        
    
if __name__=="__main__":
    import sys
    try:
        new=open(sys.argv[2],"w")
        old=open(sys.argv[1])
        dataset().fromFile(old,None).toNewFile(new)
        new.close()
        old.close()
    except:
        print "As a quick and dirty solution for transforming old-style files into new style input files you may just run:\n python data.py old_file.txt new_file.txt"
